class MacaroonException(Exception):
    pass


class MacaroonInitException(MacaroonException):
    pass


class MacaroonVerificationFailedException(MacaroonException):
    pass


class MacaroonInvalidSignatureException(MacaroonVerificationFailedException):
    pass


class MacaroonUnmetCaveatException(MacaroonVerificationFailedException):
    pass



# attempt to match encryption scheme of libmacaroons.
# currently doesn't work

# this is the first part of add_third_party_caveat

old_key = self.signature
enc_key = self._truncate_or_pad(old_key)
print('enc_key')
print(enc_key)
print(len(enc_key))
enc_nonce = self._truncate_or_pad(b'\0', size=24)
print('enc_nonce')
print(enc_nonce)
print(len(enc_nonce))
derived_key = self._truncate_or_pad(self._generate_derived_key(key))
print('derived_key')
print(derived_key)
print(len(derived_key))
enc_plaintext = self._truncate_or_pad(b'\0', size=32) + self._truncate_or_pad(derived_key, size=32)
print('enc_plaintext')
print(enc_plaintext)
print(len(enc_plaintext))
box = SecretBox(key=enc_key)
nonced_ciphertext = box.encrypt(enc_plaintext, nonce=enc_nonce)
print('nonced_ciphertext')
print(nonced_ciphertext)
print(len(nonced_ciphertext))
unnonced_ciphertext = nonced_ciphertext[24:]
print('unnonced_ciphertext')
print(unnonced_ciphertext)
print(len(unnonced_ciphertext))
enc_ciphertext = self._truncate_or_pad(unnonced_ciphertext, size=64)
print('enc_ciphertext')
print(enc_ciphertext)
print(len(enc_ciphertext))
unpadded_vid = enc_ciphertext[16:64]
print('unpadded_vid')
print(unpadded_vid)
print(len(unpadded_vid))
verificationKeyId = base64.standard_b64encode(enc_nonce + unpadded_vid)
print('verificationKeyId')
print(verificationKeyId)
print(len(verificationKeyId))

whole_cipher = base64.standard_b64encode(nonced_ciphertext)
print('whole_cipher')
print(whole_cipher)
print(len(whole_cipher))



from macaroons.utils import convert_to_string, convert_to_bytes


class Caveat(object):

    def __init__(self,
                 caveatId=None,
                 verificationKeyId=None,
                 location=None):
        self.caveatId = caveatId
        self.verificationKeyId = verificationKeyId
        self.location = location

    @property
    def caveatId(self):
        return convert_to_string(self._caveatId)

    @property
    def verificationKeyId(self):
        return convert_to_string(self._verificationKeyId)

    @property
    def location(self):
        return convert_to_string(self._location)

    @caveatId.setter
    def caveatId(self, value):
        self._caveatId = convert_to_bytes(value)

    @verificationKeyId.setter
    def verificationKeyId(self, value):
        self._verificationKeyId = convert_to_bytes(value)

    @location.setter
    def location(self, value):
        self._location = convert_to_bytes(value)

    def first_party(self):
        return self._verificationKeyId is None

    def third_party(self):
        return self._verificationKeyId is not None



from __future__ import unicode_literals

import hmac
import binascii
import struct
from hashlib import sha256
from base64 import standard_b64encode, urlsafe_b64decode, urlsafe_b64encode

from libnacl.secret import SecretBox

from macaroons.caveat import Caveat
from macaroons.utils import truncate_or_pad, convert_to_bytes


class RawMacaroon(object):
    """
    RawMacaroon uses byte types internally and in its public interface.

    This relegates all string/byte conversion and input validation to
    the to Macaroon proxy class, which is the public interface.

    RawMacaroons should not be constructed directly.

    """

    def __init__(self,
                 location=None,
                 identifier=None,
                 key=None,
                 serialized=None,
                 signature=None):
        self._serialized = serialized
        self._caveats = []
        if location and identifier and key:
            self._location = location
            self._identifier = identifier
            self._signature = self._create_initial_macaroon_signature(key)
            self._serialized = None
        elif location and identifier and signature:
            # This is only used from the verifier when we need to skip
            # the inital signature hashing for verifying
            # third party caveats
            self._location = location
            self._identifier = identifier
            self._signature = signature
            self._serialized = None
        elif serialized:
            self._serialized = serialized
            self._deserialize(serialized)
        else:
            pass

    @property
    def location(self):
        return self._location

    @property
    def identifier(self):
        return self._identifier

    @property
    def signature(self):
        return self._signature

    @property
    def caveats(self):
        return self._caveats

    # TODO
    def validate(self):
        pass

    def copy(self):
        return RawMacaroon(serialized=self.serialize())

    # Concatenates location, id, all caveats, and signature,
    # and then base64 encodes them
    def serialize(self):
        combined = self._packetize(b'location', self.location)
        combined += self._packetize(b'identifier', self.identifier)

        for caveat in self.caveats:
            combined += self._packetize(b'cid', caveat._caveatId)

            if caveat._verificationKeyId and caveat._location:
                combined += self._packetize(b'vid', caveat._verificationKeyId)
                combined += self._packetize(b'cl', caveat._location)

        combined += self._packetize(
            b'signature',
            binascii.unhexlify(self._signature)
        )
        return urlsafe_b64encode(combined)

    def serialize_json(self):
        pass

    def _deserialize(self, data):
        PACKET_PREFIX_LENGTH = 4
        decoded = urlsafe_b64decode(data)

        index = 0

        while index < len(decoded):
            packet_length = int(
                struct.unpack(
                    b"4s",
                    decoded[index:index + PACKET_PREFIX_LENGTH]
                )[0],
                16
            )
            packet = decoded[index:index + packet_length]

            start_index = index + PACKET_PREFIX_LENGTH
            end_index = index + packet_length - 1
            if packet[PACKET_PREFIX_LENGTH:].startswith(b'location'):
                self._location = \
                    decoded[start_index + len(b'location '): end_index]
                index = index + packet_length

            if packet[PACKET_PREFIX_LENGTH:].startswith(b'identifier'):
                self._identifier = \
                    decoded[start_index + len(b'identifier '): end_index]
                index = index + packet_length

            if packet[PACKET_PREFIX_LENGTH:].startswith(b'cid'):
                cid = decoded[start_index + len(b'cid '): end_index]
                self.caveats.append(Caveat(caveatId=cid))
                index = index + packet_length

            if packet[PACKET_PREFIX_LENGTH:].startswith(b'vid'):
                vid = decoded[start_index + len(b'vid '): end_index]
                self.caveats[-1].verificationKeyId = vid
                index = index + packet_length

            if packet[PACKET_PREFIX_LENGTH:].startswith(b'cl'):
                cl = decoded[start_index + len(b'cl '): end_index]
                self.caveats[-1].location = cl
                index = index + packet_length

            if packet[PACKET_PREFIX_LENGTH:].startswith(b'signature'):
                self._signature = binascii.hexlify(
                    decoded[start_index + len(b'signature '): end_index]
                )
                index = len(decoded)

        return self

    def inspect(self):
        combined = 'location' + ' ' + self.location + '\n'
        combined += 'identifier' + ' ' + self.identifier + '\n'

        for caveat in self.caveats:
            combined += 'cid' + ' ' + caveat.caveatId + '\n'

            if caveat.verificationKeyId and caveat.location:
                combined += 'vid' + ' ' + \
                    caveat.verificationKeyId.decode('ascii') + '\n'
                combined += 'cl' + ' ' + caveat.location + '\n'

        combined += 'signature' + ' ' + self.signature.decode('ascii')
        return combined

    # TODO
    def is_same(self, macaroon):
        pass

    def first_party_caveats(self):
        return [caveat for caveat in self.caveats if caveat.first_party()]

    def third_party_caveats(self):
        return [caveat for caveat in self.caveats if caveat.third_party()]

    # Protects discharge macaroons in the event they are sent to
    # the wrong location by binding to the root macaroon
    def prepare_for_request(self, macaroon):
        protected = macaroon.copy()
        protected._signature = self._macaroon_hmac_concat(
            b'\0',
            self.signature,
            macaroon.signature
        )
        return protected

    # The existing macaroon signature is the key for hashing the
    # caveat being added. This new hash becomes the signature of
    # the macaroon with caveat added.
    def add_first_party_caveat(self, predicate):
        caveat = Caveat(caveatId=predicate)
        self._caveats.append(caveat)
        encode_key = binascii.unhexlify(self.signature)
        self._signature = self._macaroon_hmac(encode_key, predicate)
        return self

    # The third party caveat key is encrypted useing the current signature, and
    # the caveat is added to the list. The existing macaroon signature
    # is the key for hashing the string (verificationId + caveatId).
    # This new hash becomes the signature of the macaroon with caveat added.
    def add_third_party_caveat(self, location, key, key_id):
        derived_key = truncate_or_pad(self._generate_derived_key(key))
        self._add_third_party_caveat_direct(
            location,
            derived_key,
            key_id,
            nonce=None
        )
        return self

    def _add_third_party_caveat_direct(self,
                                       location,
                                       key,
                                       key_id,
                                       nonce=None):
        old_key = truncate_or_pad(self.signature)
        box = SecretBox(key=old_key)
        encrypted = box.encrypt(key, nonce=nonce)
        verificationKeyId = standard_b64encode(encrypted)
        caveat = Caveat(
            caveatId=key_id,
            location=location,
            verificationKeyId=verificationKeyId
        )
        self._caveats.append(caveat)
        encode_key = binascii.unhexlify(self.signature)
        self._signature = self._macaroon_hmac_concat(
            encode_key,
            caveat._verificationKeyId,
            caveat._caveatId
        )
        return self

    # Hashes two strings, then concatenates them and hashes the combined
    # string
    def _macaroon_hmac_concat(self, key, data1, data2):
        hash1 = hmac.new(
            key,
            msg=data1,
            digestmod=sha256
        ).digest()
        hash2 = hmac.new(
            key,
            msg=data2,
            digestmod=sha256
        ).digest()
        combined = hash1 + hash2
        return hmac.new(
            key,
            msg=combined,
            digestmod=sha256
        ).hexdigest().encode('ascii')

    # Given a high-entropy root key _key and an identifier id, this returns
    # a valid signature sig = MAC(k, id).
    def _create_initial_macaroon_signature(self, key):
        derived_key = self._generate_derived_key(key)
        return self._macaroon_hmac(derived_key, self._identifier)

    def _generate_derived_key(self, key):
        generator_key = b'macaroons-key-generator'
        derived_key = self._hmac(generator_key, key)
        return derived_key

    def _macaroon_hmac(self, key, data):
        dig = self._hmac(key, data)
        return binascii.hexlify(dig)

    def _hmac(self, key, data):
        return hmac.new(
            key,
            msg=data,
            digestmod=sha256
        ).digest()

    def _packetize(self, key, data):
        PACKET_PREFIX_LENGTH = 4
        # The 2 covers the space and the newline
        packet_size = PACKET_PREFIX_LENGTH + 2 + len(key) + len(data)
        # Ignore the first two chars, 0x
        packet_size_hex = hex(packet_size)[2:]
        header = packet_size_hex.zfill(4).encode('ascii')
        packet_content = key + b' ' + data + b'\n'
        packet = struct.pack(
            convert_to_bytes("4s%ds" % len(packet_content)),
            header,
            packet_content
        )
        return packet



from six import text_type, binary_type


def convert_to_bytes(string):
    if string is None:
        return None
    if type(string) is text_type:
        return string.encode('ascii')
    elif type(string) is binary_type:
        return string
    else:
        raise TypeError("Must be a string or bytes object.")


def convert_to_string(bytes):
    if bytes is None:
        return None
    if type(bytes) is text_type:
        return bytes
    elif type(bytes) is binary_type:
        return bytes.decode('ascii')
    else:
        raise TypeError("Must be a string or bytes object.")


def truncate_or_pad(byte_string, size=None):
    if size is None:
        size = 32
    byte_array = bytearray(byte_string)
    length = len(byte_array)
    if length > size:
        return bytes(byte_array[:size])
    elif length < size:
        return bytes(byte_array + b"\0"*(size-length))
    else:
        return byte_string


def equals(val1, val2):
    """
    Returns True if the two strings are equal, False otherwise.

    The time taken is independent of the number of characters that match.

    For the sake of simplicity, this function executes in constant time only
    when the two strings have the same length. It short-circuits when they
    have different lengths.
    """
    if len(val1) != len(val2):
        return False
    result = 0
    for x, y in zip(val1, val2):
        result |= ord(x) ^ ord(y)
    return result == 0
